#'Compute the Ranked Probability Score
#'
#'The Ranked Probability Score (RPS; Wilks, 2011) is defined as the sum of the
#'squared differences between the cumulative forecast probabilities (computed 
#'from the ensemble members) and the observations (defined as 0% if the category
#'did not happen and 100% if it happened). It can be used to evaluate the skill 
#'of multi-categorical probabilistic forecasts. The RPS ranges between 0 
#'(perfect forecast) and n-1 (worst possible forecast), where n is the number of
#'categories. In the case of a forecast divided into two categories (the lowest
#'number of categories that a probabilistic forecast can have), the RPS 
#'corresponds to the Brier Score (BS; Wilks, 2011), therefore, ranges between 0 
#'and 1.
#'
#'@param exp A named numerical array of the forecast with at least time 
#'  dimension.  
#'@param obs A named numerical array of the observation with at least time 
#'  dimension. The dimensions must be the same as 'exp' expect 'memb_dim'.
#'@param time_dim A character string indicating the name of the time dimension.
#'  The default value is 'sdate'.
#'@param memb_dim A character string indicating the name of the member dimension
#'  to compute the ensemble mean of the forecast. If it is NULL, the ensemble
#'  mean should be provided directly to the function. The default value is NULL.
#'@param prob_thresholds A numeric vector of the relative thresholds (from 0 to
#'  1) between the categories. The default value is c(1/3, 2/3), which 
#'  corresponds to tercile equiprobable categories.
#'@param indices_for_clim A vector of the indices to be taken along "time_dim" 
#'  for computing the thresholds between the probabilistic categories. If NULL,
#'  the whole period is used. The default value is NULL.
#'@param Fair A logical indicating whether to compute the FairRPSS (the 
#'  potential RPSS that the forecast would have with an infinite ensemble size).
#'  The default value is FALSE.
#'@param ncores An integer indicating the number of cores to use for parallel 
#'  computation. The default value is NULL.
#'
#'@return 
#'A numerical array of RPS with the same dimensions as "exp" except the 
#'"time_dim" and "memb_dim" dimensions.
#'
#'@references 
#'Wilks, 2011; https://doi.org/10.1016/B978-0-12-385022-5.00008-7
#'
#'@examples
#' exp <- array(rnorm(1000), dim = c(lat = 3, lon = 2, member = 10, sdate = 50))
#' obs <- array(rnorm(1000), dim = c(lat = 3, lon = 2, sdate = 50))
#' res <- RPS(exp = exp, obs = obs, memb_dim = 'member')
#'
#'@import multiApply
#'@importFrom ClimProjDiags Subset
#'@importFrom easyVerification convert2prob
#'@export
RPS <- function(exp, obs, time_dim = 'sdate', memb_dim = NULL, 
                prob_thresholds = c(1/3, 2/3), indices_for_clim = NULL, Fair = FALSE, 
                ncores = NULL) {
  
  # Check inputs
  ## exp and obs (1)

  if (!is.array(exp)){stop('"exp" must be an array.')}
  if (!is.array(obs)){stop('"obs" must be an array.')}
  if (!identical(as.numeric(dim(exp)[time_dim]), as.numeric(dim(obs)[time_dim]))){stop('"time_dim" must have the same length in "exp" and "obs".')}
  if (!is.numeric(prob_thresholds) | !is.vector(prob_thresholds) | any(prob_thresholds <= 0) | any(prob_thresholds >= 1)){
    stop('"prob_thresholds" must be a numeric vector with the relative thresholds (from 0 to 1) between the categories.')
  }
  if (is.null(indices_for_clim)){
    indices_for_clim <- 1:as.numeric(dim(obs)[time_dim])
  } else if (!is.numeric(indices_for_clim)){
    stop('"indices_for_clim" must be NULL or a numeric vector.')
  }
  if (!is.logical(Fair)){stop('"Fair" must be TRUE or FALSE.')}
  if (!is.character(time_dim)){stop('"time_dim" must be a string.')}
  if (!is.character(memb_dim)){stop('"memb_dim" must be a string.')}
  if (!is.null(ncores)){
    if (!is.numeric(ncores) | ncores < 1){stop('"ncores" must be either NULL or a positive integer.')}
  }
  
  ## RPS
  rps <- Apply(data = list(exp = exp, obs = obs), 
               target_dims = list(exp = c(time_dim, memb_dim), obs = time_dim),
               fun = .RPS,
               prob_thresholds = prob_thresholds, indices_for_clim = indices_for_clim,
               Fair = Fair,
               ncores = ncores)$output1
  
  return(rps)
}


.RPS <- function(exp, obs, prob_thresholds, indices_for_clim, Fair) {
  # exp: [sdate, memb]
  # obs: [sdate]

  exp_probs <- .get_probs(data = exp, indices_for_quantiles = indices_for_clim, 
                          prob_thresholds = prob_thresholds)
  # exp_probs: [bin, sdate] 
  obs_probs <- .get_probs(data = obs, indices_for_quantiles = indices_for_clim, 
                          prob_thresholds = prob_thresholds)
  # obs_probs: [bin, sdate]
 
  ## RPS for each time step
  rps <- Apply(data = list(probs_exp = exp_probs, probs_obs = obs_probs), target_dims = 'bin',
                           fun = .RPS_from_probs, Fair = Fair, 
                           n_members = dim(exp)[2], ncores = NULL)$output1
  # rps: [sdate]

  # Return only the mean RPS
  rps <- mean(rps, na.rm = FALSE)

  return(rps)
}


.get_probs <- function(data, indices_for_quantiles, prob_thresholds) {
  # if exp: [sdate, memb]
  # if obs: [sdate]
  if (length(dim(data)) == 1) dim(data) <- c(dim(data), 1)

  # Absolute thresholds
  quantiles <- quantile(x = as.vector(data), probs = prob_thresholds, type = 8, na.rm = TRUE)

  # Probabilities
  probs <- array(dim = c(bin = length(quantiles) + 1, dim(data)[1])) # [bin, sdate]
  for (i_time in 1:dim(data)[1]) {
    if (anyNA(data[i_time, ])) {
    probs[, i_time] <- rep(NA, dim = length(quantiles) + 1)
    } else {
    probs[, i_time] <- colMeans(easyVerification::convert2prob(data[i_time, ], threshold = quantiles))
    }
  }

  return(probs)
}


.RPS_from_probs <- function(probs_exp, probs_obs, Fair, n_members) {
  # probs_exp: [bin]
  # probs_obs: [bin]

  rps <- sum((cumsum(probs_exp) - cumsum(probs_obs))^2)

  if (Fair) { # FairRPS
    ## adjustment <- rowSums(-1 * (1/R - 1/R.new) * ens.cum * (R - ens.cum)/R/(R - 1)) [formula taken from SpecsVerification::EnsRps]
    R <- n_members
    R_new <- Inf 
    probs_cum <- cumsum(probs_exp)
    adjustment <- -1 / (R - 1) * probs_cum * (1 - probs_cum)
    adjustment <- sum(adjustment)
    rps <- rps + adjustment
  }
  
  return(rps)
}


